@@ -232,13 +232,8 @@ class Source(object):
             for node in ast.walk(self.tree):
                 for child in ast.iter_child_nodes(node):
                     child.parent = node
-                if hasattr(node, "lineno"):
-                    if hasattr(node, "end_lineno") and isinstance(node, ast.expr):
-                        linenos = range(node.lineno, node.end_lineno + 1)
-                    else:
-                        linenos = [node.lineno]
-                    for lineno in linenos:
-                        self._nodes_by_line[lineno].append(node)
+                for lineno in node_linenos(node):
+                    self._nodes_by_line[lineno].append(node)
 
             visitor = QualnameVisitor()
             visitor.visit(self.tree)
@@ -252,23 +247,32 @@ class Source(object):
         return cls.for_filename(frame.f_code.co_filename, frame.f_globals or {}, use_cache)
 
     @classmethod
-    def for_filename(cls, filename, module_globals=None, use_cache=True):
+    def for_filename(
+        cls,
+        filename,
+        module_globals=None,
+        use_cache=True,  # noqa no longer used
+    ):
         if isinstance(filename, Path):
             filename = str(filename)
 
-        source_cache = cls._class_local('__source_cache', {})
-        if use_cache:
-            try:
-                return source_cache[filename]
-            except KeyError:
-                pass
+        def get_lines():
+            return linecache.getlines(filename, module_globals)
 
-        if not use_cache:
-            linecache.checkcache(filename)
+        # Save the current linecache entry, then ensure the cache is up to date.
+        entry = linecache.cache.get(filename)
+        linecache.checkcache(filename)
+        lines = get_lines()
+        if entry is not None and not lines:
+            # There was an entry, checkcache removed it, and nothing replaced it.
+            # This means the file wasn't simply changed (because the `lines` wouldn't be empty)
+            # but rather the file was found not to exist, probably because `filename` was fake.
+            # Restore the original entry so that we still have something.
+            linecache.cache[filename] = entry
+            lines = get_lines()
 
-        lines = tuple(linecache.getlines(filename, module_globals))
-        result = source_cache[filename] = cls._for_filename_and_lines(filename, lines)
-        return result
+        lines = tuple(lines)
+        return cls._for_filename_and_lines(filename, lines)
 
     @classmethod
     def _for_filename_and_lines(cls, filename, lines):
@@ -313,61 +317,31 @@ class Source(object):
         key = (code, id(code), lasti)
         executing_cache = cls._class_local('__executing_cache', {})
 
-        try:
-            args = executing_cache[key]
-        except KeyError:
-            def find(source, retry_cache):
-                node = stmts = decorator = None
-                tree = source.tree
-                if tree:
-                    try:
-                        stmts = source.statements_at_line(lineno)
-                        if stmts:
-                            if is_ipython_cell_code(code):
-                                for stmt in stmts:
-                                    tree = _extract_ipython_statement(stmt)
-                                    try:
-                                        node_finder = NodeFinder(frame, stmts, tree, lasti)
-                                        if (node or decorator) and (node_finder.result or node_finder.decorator):
-                                            if retry_cache:
-                                                raise AssertionError
-                                            # Found potential nodes in separate statements,
-                                            # cannot resolve ambiguity, give up here
-                                            node = decorator = None
-                                            break
+        args = executing_cache.get(key)
+        if not args:
+            node = stmts = decorator = None
+            source = cls.for_frame(frame)
+            tree = source.tree
+            if tree:
+                try:
+                    stmts = source.statements_at_line(lineno)
+                    if stmts:
+                        if is_ipython_cell_code(code):
+                            decorator, node = find_node_ipython(frame, lasti, stmts)
+                        else:
+                            node_finder = NodeFinder(frame, stmts, tree, lasti)
+                            node = node_finder.result
+                            decorator = node_finder.decorator
+                except Exception:
+                    if TESTING:
+                        raise
 
-                                        node = node_finder.result
-                                        decorator = node_finder.decorator
-                                    except Exception:
-                                        if retry_cache:
-                                            raise
+                if node:
+                    new_stmts = {statement_containing_node(node)}
+                    assert_(new_stmts <= stmts)
+                    stmts = new_stmts
 
-                            else:
-                                node_finder = NodeFinder(frame, stmts, tree, lasti)
-                                node = node_finder.result
-                                decorator = node_finder.decorator
-                    except Exception as e:
-                        # These exceptions can be caused by the source code having changed
-                        # so the cached Source doesn't match the running code
-                        # (e.g. when using IPython %autoreload)
-                        # Try again with a fresh Source object
-                        if retry_cache and isinstance(e, (NotOneValueFound, AssertionError)):
-                            return find(
-                                source=cls.for_frame(frame, use_cache=False),
-                                retry_cache=False,
-                            )
-                        if TESTING:
-                            raise
-
-                    if node:
-                        new_stmts = {statement_containing_node(node)}
-                        assert_(new_stmts <= stmts)
-                        stmts = new_stmts
-
-                return source, node, stmts, decorator
-
-            args = find(source=cls.for_frame(frame), retry_cache=True)
-            executing_cache[key] = args
+            executing_cache[key] = args = source, node, stmts, decorator
 
         return Executing(frame, *args)
 
@@ -576,11 +550,13 @@ class NodeFinder(object):
         self.instruction = instruction = self.get_actual_current_instruction(lasti)
         op_name = instruction.opname
         extra_filter = lambda e: True
+        ctx = type(None)
 
         if op_name.startswith('CALL_'):
             typ = ast.Call
         elif op_name.startswith(('BINARY_SUBSCR', 'SLICE+')):
             typ = ast.Subscript
+            ctx = ast.Load
         elif op_name.startswith('BINARY_'):
             typ = ast.BinOp
             op_type = dict(
@@ -610,15 +586,23 @@ class NodeFinder(object):
             extra_filter = lambda e: isinstance(e.op, op_type)
         elif op_name in ('LOAD_ATTR', 'LOAD_METHOD', 'LOOKUP_METHOD'):
             typ = ast.Attribute
-            # `in` to handle private mangled attributes
-            extra_filter = lambda e: e.attr in instruction.argval
+            ctx = ast.Load
+            extra_filter = lambda e: attr_names_match(e.attr, instruction.argval)
         elif op_name in ('LOAD_NAME', 'LOAD_GLOBAL', 'LOAD_FAST', 'LOAD_DEREF', 'LOAD_CLASSDEREF'):
             typ = ast.Name
+            ctx = ast.Load
             if PY3 or instruction.argval:
                 extra_filter = lambda e: e.id == instruction.argval
         elif op_name in ('COMPARE_OP', 'IS_OP', 'CONTAINS_OP'):
             typ = ast.Compare
             extra_filter = lambda e: len(e.ops) == 1
+        elif op_name.startswith(('STORE_SLICE', 'STORE_SUBSCR')):
+            ctx = ast.Store
+            typ = ast.Subscript
+        elif op_name.startswith('STORE_ATTR'):
+            ctx = ast.Store
+            typ = ast.Attribute
+            extra_filter = lambda e: attr_names_match(e.attr, instruction.argval)
         else:
             raise RuntimeError(op_name)
 
@@ -628,11 +612,18 @@ class NodeFinder(object):
                 for stmt in stmts
                 for node in ast.walk(stmt)
                 if isinstance(node, typ)
-                if not (hasattr(node, "ctx") and not isinstance(node.ctx, ast.Load))
+                if isinstance(getattr(node, "ctx", None), ctx)
                 if extra_filter(node)
                 if statement_containing_node(node) == stmt
             }
 
+            if ctx == ast.Store:
+                # No special bytecode tricks here.
+                # We can handle multiple assigned attributes with different names,
+                # but only one assigned subscript.
+                self.result = only(exprs)
+                return
+
             matching = list(self.matching_nodes(exprs))
             if not matching and typ == ast.Call:
                 self.find_decorator(stmts)
@@ -1052,10 +1043,10 @@ def assert_linenos(tree):
     for node in ast.walk(tree):
         if (
                 hasattr(node, 'parent') and
-                hasattr(node, 'lineno') and
                 isinstance(statement_containing_node(node), ast.Assert)
         ):
-            yield node.lineno
+            for lineno in node_linenos(node):
+                yield lineno
 
 
 def _extract_ipython_statement(stmt):
@@ -1086,3 +1077,44 @@ def is_ipython_cell_code(code_obj):
         is_ipython_cell_filename(code_obj.co_filename) and
         is_ipython_cell_code_name(code_obj.co_name)
     )
+
+
+def find_node_ipython(frame, lasti, stmts):
+    node = decorator = None
+    for stmt in stmts:
+        tree = _extract_ipython_statement(stmt)
+        try:
+            node_finder = NodeFinder(frame, stmts, tree, lasti)
+            if (node or decorator) and (node_finder.result or node_finder.decorator):
+                # Found potential nodes in separate statements,
+                # cannot resolve ambiguity, give up here
+                return None, None
+
+            node = node_finder.result
+            decorator = node_finder.decorator
+        except Exception:
+            pass
+    return decorator, node
+
+
+def attr_names_match(attr, argval):
+    """
+    Checks that the user-visible attr (from ast) can correspond to
+    the argval in the bytecode, i.e. the real attribute fetched internally,
+    which may be mangled for private attributes.
+    """
+    if attr == argval:
+        return True
+    if not attr.startswith("__"):
+        return False
+    return bool(re.match(r"^_\w+%s$" % attr, argval))
+
+
+def node_linenos(node):
+    if hasattr(node, "lineno"):
+        if hasattr(node, "end_lineno") and isinstance(node, ast.expr):
+            linenos = range(node.lineno, node.end_lineno + 1)
+        else:
+            linenos = [node.lineno]
+        for lineno in linenos:
+            yield lineno
